
//-------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <time.h>

//-------------------------------------------------------------

FILE *fp, *fz, *fk;		// File in, out, and key file
unsigned char *src;		// source
unsigned char *key;		// key
unsigned int nLen;		// source length
unsigned int kLen;		// key length
unsigned int kAnd;		// key length (for & use, much faster than modulo)

unsigned char xor1;		// xor 1
unsigned int  xor2;		// xor 2
unsigned char tabs[4][0x100];	// tables
unsigned char *tab_a, *tab_b;	// pointer to tables
unsigned char pTaba, pTabb;	// block xor

//-------------------------------------------------------------

// xor m & n by 1 byte from table 'a' and table 'b' respectively.
// set which bit is used in a horribly convoluded way.
#define _xor1_0(j, k) \
			m ^= tab_a[tab_b[m ^ key[(((i + m) ^ j) ^ xor1) & kAnd]] ^ n] & (0x80 >> k); \
			n ^= tab_b[tab_a[n ^ key[(((i + n) ^ j) ^ xor1) & kAnd]] ^ m] & (0x01 << k); \

#define _xor1(j) \
			_xor1_0((j) ^ (xor1 & 7), (j))

// xor by 8 bits of i (the address) set by tables
#define ixor(tab0, tab1) \
			xor1 ^= i >> (((tab0[i & n] >> 4) & 5) | (tab1[m] & 2)) \

// rotate xor1's bits around a bit
#define rotate(j) \
			xor1 = (xor1 << j) | (xor1 >> (8 - j)) \

// reverse order of the bits in an 8 bit int
#define rev(n) (((n & 1) << 7) | ((n & 2) << 5) | ((n & 4) << 3) | ((n & 8) << 1) \
		| ((n >> 1) & 8) | ((n >> 3) & 4) | ((n >> 5) & 2) | ((n >> 7) & 1))

// put pairs together (1, 0x10, 2, 0x20, 4, 0x40...)
#define swr(n) ((n & 0x81) | ((n >> 3) & 2) | ((n & 2) << 1) | ((n >> 2) & 8) | ((n & 4) << 2) | ((n >> 1) & 0x20) | ((n & 8) << 3))

// xor a single bit set by j, xor1, and i
#define invert_bit(j) (1 << (j ^ ((xor1 ^ i) & 0x07)))

//-------------------------------------------------------------

unsigned char byte_xor2(unsigned short ofst, unsigned char i, unsigned int c)
{
	unsigned char r;

	// update seed
	xor2 <<= 0x01;
	xor2 |= (key[xor2 % kLen] >> (xor1 & 7)) & 1;
	xor2 ^= c;
	xor2 ^= i << 0x04;
	xor2 ^= rev(ofst >> 0x04) << 0x18;
	xor2 ^= ofst & 0x00f;

	// use highest byte to seed returned byte
	r = xor2 >> 0x18;

	// use seed to modify the returned value
	if ( xor2 & 0x00000008) r  = (r & 0xfe) | (xor2 & 0x01);
	if ( xor2 & 0x00000004) r  = (r << 0x01) | (r >> 0x07);
	if ( xor2 & 0x00000002) r ^= (xor2 >> 0x08) & 0x09;
	if (~xor2 & 0x00000001) r  = (r & 0xbf) | ((xor2 >> (i & 0x10)) & 0x40);
	if ( xor2 & 0x00000080) r  = invert_bit(7);
	if ( xor2 & 0x00000040) r  = (r << 0x04) | (r >> 0x04);
	if ( xor2 & 0x00000020) r ^= (xor2 >> 0x08) & 0x60;
	if ( xor2 & 0x00000010) r  = (r & 0xfb) | (xor2 & 0x04);
	if ( xor2 & 0x00000800) r ^= invert_bit(4);
	if (~xor2 & 0x00000400) r  = rev(r);
	if ( xor2 & 0x00000200) r ^= invert_bit(2);
	if (~xor2 & 0x00000100) r ^= (xor2 >> 0x08) & 0x90;
	if ( xor2 & 0x00008000) r  = (r & 0xf7) | (xor2 & 0x08);
	if ( xor2 & 0x00004000) r  = swr(r);
	if ( xor2 & 0x00002000) r ^= (xor2 >> 0x08) & 0x06;
	if ( xor2 & 0x00001000) r ^= invert_bit(6);
	if ( xor2 & 0x00080000) r  = (r & 0xef) | ((xor2 >> (c & 0x10)) & 0x10);
	if ( xor2 & 0x00040000) r  = (r << 0x07) | (r >> 0x01);
	if ( xor2 & 0x00020000) r ^= xor2 & 0xc0;
	if (~xor2 & 0x00010000) r ^= invert_bit(1);
	if (~xor2 & 0x00800000) r  = (r & 0xdf) | ((xor2 >> (c & 0x10)) & 0x20);
	if ( xor2 & 0x00400000) r ^= invert_bit(3);
	if ( xor2 & 0x00200000) r ^= xor2 & 0x03;
	if ( xor2 & 0x00100000) r ^= xor2 & 0x30;
	if (~xor2 & 0x08000000) r ^= invert_bit(0);
	if ( xor2 & 0x04000000) r  = (r & 0xfc) | ((r & 0x01) << 0x01) | ((r & 0x02) >> 0x01);
	if ( xor2 & 0x02000000) r ^= invert_bit(5);
	if (~xor2 & 0x01000000) r  = (r & 0xf9) | ((r & 0x02) << 0x01) | ((r & 0x04) >> 0x01);
	if ( xor2 & 0x80000000) r  = (r & 0x7f) | ((xor2 >> (i & 0x10)) & 0x80);
	if ( xor2 & 0x40000000) r ^= ~rev(r);
	if ( xor2 & 0x20000000) r ^= xor2 & 0x0c;
	if ( xor2 & 0x10000000) r ^= (r & 0xfd) | (xor2 & 0x02);

	// modify the seed using the returned value
	xor2 ^= r;

	return r;
}

//-------------------------------------------------------------

void fill_tab(unsigned char *tab, unsigned short ofst)
{
	int i = 0, j, c = 0;

	memset (tab, xor2 & 0xff, 0x100);					// set the entire table to a single value

	while (i < 0xff)							// do not modify tab[0xff] (inf. loop otherwise)
	{
		unsigned char d = byte_xor2(ofst, i, c);			// get value, offset, pos in table, total tries

		j = 0;
		while (tab[j] != d && j < 0x100) j++;				// check current table for this value

		if (j == 0x100) {						// if it is not found
			tab[i] = d;						// add it
			i++;							// and move on to the next position
		}

		c++;								// increase counter
	}

	pTaba ^= c;
	pTabb ^= byte_xor2(*((unsigned short *)(tab + 0)), tab[2], *((unsigned int *)(tab + 4)));
}

void fill_tabs(unsigned short ofst)						// fill tables using xor2 routine
{
	pTaba = pTabb = 0;
	fill_tab(tabs[0 ^ (key[ofst % kLen] & 3)], ofst);			// fill each table depending on the offset
	fill_tab(tabs[1 ^ (key[ofst % kLen] & 3)], ofst);
	fill_tab(tabs[2 ^ (key[ofst % kLen] & 3)], ofst);
	fill_tab(tabs[3 ^ (key[ofst % kLen] & 3)], ofst);
}

//-------------------------------------------------------------


unsigned char byte_xor1(unsigned int i)
{
	unsigned char m, n;

	m = i >> 0x18;								// seed m
	n = i >> 0x10;								// seed n

	_xor1(0);								// xor m, n
	_xor1(1);
	_xor1(2);
	_xor1(3);
	_xor1(4);
	_xor1(5);
	_xor1(6);
	_xor1(7);

	tab_a = tabs[((i ^ n) & 2)    ]		;				// set tables again, this way the data doesn't
	tab_b = tabs[((i ^ m) & 2) | 1];					// have to be xored by the same table as m & n

	rotate(m & 7);								// rotate

	ixor(tab_a, tab_b);							// xor using address
	ixor(tab_b, tab_a);

	rotate(n & 7);								// rotate

	xor1 ^= tab_a[xor1 ^ key[xor2 % kLen]];					// xor using table 'a'
	xor1 ^= tab_b[i % ((m ^ n) + 1)];					// xor using table 'b'

	return xor1;
}

//-------------------------------------------------------------

void crypt(unsigned char *src, unsigned int nLen)
{
	unsigned int i, j;
	unsigned int block = 0;
	
	while (kAnd > kLen) kAnd >>= 1;						// set kAnd

	tab_a = tabs[(key[0] << 1) & 2];					// set tables
	tab_b = tabs[(key[1] & 2) | 1];

	xor1 = (key[0] << 4) | (key[1] & 0x0f);					// seed data xor 1
	xor2 = key[xor1 % kLen] ^ (key[~xor1 % kLen] << 4);			// seed data xor 2

	for (i = 0; i < nLen; i+= 0x100000)
	{
		fill_tabs(i >> 20);						// fill tables every block

		block = (nLen - i) < 0x100000 ? (nLen - i) : 0x100000;		// set block size (Max or what's left)

		fread (src, 1, block, fp);					// read block

		for (j = 0; j < block; j++)					// apply encryption to every byte
		{
			src[j] ^= byte_xor1(i + j);				// data xor 1
			src[j] ^= byte_xor2(i >> 20, xor1, j);			// data xor 2
			src[j] ^= tab_a[pTaba] ^ tab_b[pTabb];			// data xor 3 (block xor)
		}

		fwrite (src, 1, block, fz);					// write block
	}

	memset (key, 0, kLen);							// Clear the key block
	memset (tabs, 0, 0x400);						// Clear the tabs
	kLen = kAnd = xor1 = xor2;						// Clear vars
}


//-------------------------------------------------------------

int main(int argc, char **argv)
{
	char filein[256], fileout[256];
	char tkey[256];

	kAnd = 0xFFFFFFFF;

	if (argc < 2)
	{
		printf ("Please type the name of the file to be en/de-crypted\n");
		//gets (filein);
		fgets(filein, 256, stdin);
		if (filein[strlen(filein)-1] == '\n') filein[strlen(filein)-1] = '\0';
		strcpy (fileout, filein);
	}
	else
	{
		strcpy (fileout, argv[1]);
	}

	fp = fopen(fileout, "rb");

	if (fp == NULL)
	{
		printf ("File not found\n");
		fgets (filein, 1, stdin);
		return 1;
	}

	fseek (fp, 0, SEEK_END);
	nLen = ftell (fp);
	rewind (fp);

	if (argc < 3)
	{
		printf ("Please type the name of the key file or the password you wish to use.\n");
		//gets (tkey);
		fgets(tkey, 256, stdin);
		if (tkey[strlen(tkey)-1] == '\n') tkey[strlen(tkey)-1] = '\0';

	}
	else
	{
		strcpy (tkey, argv[2]);
	}

	fk = fopen(tkey, "rb");
	if (fk != NULL)
	{
		fseek (fk, 0, SEEK_END);
		kLen = ftell (fk);
		rewind (fk);

		key = (unsigned char *)malloc ( kLen );

		fread (key, 1, kLen, fk);
		fclose (fk);
	}
	else
	{
		key = tkey;
		kLen = strlen( tkey );
	}

	strcat (fileout, ".enc");

	fz = fopen(fileout, "wb");

	src = (unsigned char*)malloc( 0x100000 );

time_t xx = time(NULL);
	crypt(src, nLen);
printf ("%d\n", time(NULL)-xx);
	fclose (fp);
	fclose (fz);

	free (key);
	free (src);
}
